-- | THE constant pool, all the constants in a class file are handled in here.

{-# LANGUAGE DerivingStrategies #-}
{-# LANGUAGE InstanceSigs #-}
{-# LANGUAGE TypeApplications #-}
{-# LANGUAGE OverloadedStrings #-}
{-# LANGUAGE StandaloneKindSignatures #-}
{-# LANGUAGE DataKinds #-}
{-# LANGUAGE TypeFamilies #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE GADTs #-}
{-# LANGUAGE StandaloneDeriving #-}
module Language.Java.Classfile.ConstantPool (ConstantPool(..)) where
import Data.Word (Word16)
import Data.Array.IArray (Array, listArray)
import Language.Java.Classfile.ConstantPool.Entry (Entry, StorageCount (..), storageCount)
import Language.Java.Classfile.Extractable (Extractable (extract))
import Language.Java.Classfile.Extract (Extract, traceIndex, traceType)
import qualified Data.Text as Text
import Language.Java.Classfile.Stage (Stage(..))
import Data.Kind (Type)

type ConstantPool :: Stage -> Type
data ConstantPool stage where
  ConstantPool :: (Array Word16 Entry) -> ConstantPool Parse
  NoPool :: ConstantPool Resolve

deriving instance Show (ConstantPool stage)

-- | 'Stage'-indexed constant-pool. The constant-pool is erased after resolving the class file.


instance Extractable (ConstantPool Parse) where
  extract :: Extract (ConstantPool Parse)
  extract = do
    count <- extract @Word16

    let typeName = Text.concat
          [ "ConstantPool ("
          , Text.pack . show $ count
          , ")"
          ]

    let extractElements index end
          | index == end = pure []
          | otherwise = do
            element <- traceIndex index extract
            case storageCount element of
              Once  -> (element:) <$> extractElements (succ index) end
              Twice -> ([element, element] <>) <$> extractElements (index + 2) end

    traceType typeName $ do
      elements <- extractElements 1 (fromIntegral count)
      pure . ConstantPool $ listArray (1, count - 1) elements
