-- | Collection of types used in the classfile, they are generally an index into the constant-pool.

{-# LANGUAGE GeneralizedNewtypeDeriving #-}
{-# LANGUAGE DataKinds #-}
{-# LANGUAGE DerivingVia #-}
{-# LANGUAGE TypeFamilies #-}
{-# LANGUAGE InstanceSigs #-}
{-# LANGUAGE LambdaCase #-}
{-# LANGUAGE StandaloneKindSignatures #-}
module Language.Java.Classfile.ConstantPool.References (Utf8Reference(..), ClassReference(..), NameAndTypeReference(..), MethodHandleReferenceKind(..), OpaqueReference(..), BootstrapMethodIndex(..), Class) where
import Data.Word (Word16, Word8)
import Language.Java.Classfile.Extractable (Extractable)
import Language.Java.Classfile.Extractable.AsTag
    ( TagValue(..), AsTag, AsTag(..) )
import Language.Java.Classfile.Stage (Stage(..))
import Data.Kind (Type)

-- | Wrapper for constant-pool reference to text.

newtype Utf8Reference = Utf8Reference Word16
  deriving stock (Show)
  deriving newtype Extractable

-- | 'Stage'-indexed type, either a Class or only a t'ClassReference'.

type Class :: Stage -> Type
type family Class stage where
  Class Parse = ClassReference

-- | Reference to a class in a constant-pool. This will resolve into a class.

newtype ClassReference = ClassReference Word16
  deriving stock (Show)
  deriving newtype Extractable

-- | Reference to a class in a constant-pool. This will resolve to Name and Type.

newtype NameAndTypeReference = NameAndTypeReference Word16
  deriving stock (Show)
  deriving newtype Extractable

-- | Reference to something in a constant-pool. I will hopefully get rid of this type.

newtype OpaqueReference = OpaqueReference Word16
  deriving stock (Show)
  deriving newtype Extractable

-- | Reference to a method in the BootstrapMethods class attribute

newtype BootstrapMethodIndex = BootstrapMethodIndex Word16
  deriving stock (Show)
  deriving newtype Extractable

-- | A Tag used to determine the type of a MethodHandle

data MethodHandleReferenceKind
  = GetField
  | GetStatic
  | PutField
  | PutStatic
  | InvokeVirtual
  | InvokeStatic
  | InvokeSpecial
  | NewInvokeSpecial
  | InvokeInterface
  deriving stock (Show, Enum, Bounded)
  deriving Extractable via AsTag MethodHandleReferenceKind

instance TagValue MethodHandleReferenceKind where
  type TagType MethodHandleReferenceKind = Word8
  tagOf :: MethodHandleReferenceKind -> TagType MethodHandleReferenceKind
  tagOf = \case
    GetField         -> 1
    GetStatic        -> 2
    PutField         -> 3
    PutStatic        -> 4
    InvokeVirtual    -> 5
    InvokeStatic     -> 6
    InvokeSpecial    -> 7
    NewInvokeSpecial -> 8
    InvokeInterface  -> 9
