-- | Treat some enumeration value as a tag. Use for deriving via.

{-# LANGUAGE TypeFamilies #-}
{-# LANGUAGE InstanceSigs #-}
{-# LANGUAGE TypeApplications #-}
{-# LANGUAGE ScopedTypeVariables #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE UndecidableInstances #-}
module Language.Java.Classfile.Extractable.AsTag (AsTag(..), TagValue(..)) where
import Data.Kind (Type)
import Language.Java.Classfile.Extractable (Extractable (extract), expectConstant)
import Language.Java.Classfile.Extract (Extract)
import Data.Enum.Util (enumerate)
import Control.Applicative (asum)
import Data.Functor (($>))
import Data.Typeable ( Typeable )

-- | The 'Extractable' instance will use the values from 'TagValue' to get the correct constructor.

newtype AsTag a = AsTag a
  deriving Show

instance (Extractable (TagType a), Bounded a, Enum a, Eq (TagType a), Show (TagType a), Typeable (TagType a), TagValue a ) => Extractable (AsTag a) where
  extract :: Extract (AsTag a)
  extract = do
    let expectTag con = expectConstant (tagOf con) $> con
    let allConstructors = expectTag <$> enumerate @a

    AsTag <$> asum allConstructors


-- | Specify the type of a tag. Provide values using 'tagOf'. This relies on the 'Enum' and 'Bounded' instances of the wrapped type to extract them.

class TagValue a where
  type TagType a :: Type
  tagOf :: a -> TagType a
