-- | Fail if a Value doesn't have the right tag.

{-# LANGUAGE DerivingStrategies #-}
{-# LANGUAGE GeneralizedNewtypeDeriving #-}
{-# LANGUAGE InstanceSigs #-}
{-# LANGUAGE StandaloneKindSignatures #-}
{-# LANGUAGE TypeApplications #-}
{-# LANGUAGE ScopedTypeVariables #-}
{-# LANGUAGE DataKinds #-}
module Language.Java.Classfile.Extractable.WithTag (Word8Tag, WithNumericTag(..)) where
import Language.Java.Classfile.Extractable (Extractable (extract), expectConstant)
import Language.Java.Classfile.Extract (Extract)
import GHC.TypeLits (Natural, natVal, KnownNat)
import Data.Kind (Type)
import Control.Monad (void)
import Data.Proxy (Proxy(Proxy))
import Data.Typeable ( Typeable )
import Data.Word (Word8)

-- | Type alias if you use the same type a lot.

type Word8Tag value a = WithNumericTag value Word8 a

-- | This uses 'DataKinds' to extract the type-level natural number. Negative Tags are not supported.

type WithNumericTag :: Natural -> Type -> Type -> Type
newtype WithNumericTag value tagType a = Tagged a
  deriving newtype Show

instance (KnownNat value, Extractable a, Extractable tagType, Eq tagType, Num tagType, Show tagType, Typeable tagType) => Extractable (WithNumericTag value tagType a) where
  extract :: Extract (WithNumericTag value tagType a)
  extract = do
    void $ expectConstant @tagType (fromIntegral . natVal $ Proxy @value)
    Tagged <$> extract

