-- | Parse flags, this will produce a Set of specified flags.

{-# LANGUAGE InstanceSigs #-}
{-# LANGUAGE TypeApplications #-}
{-# LANGUAGE ScopedTypeVariables #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE UndecidableInstances #-}
{-# LANGUAGE TypeFamilies #-}
module Language.Java.Classfile.Flags (Flags(..), FlagMask(..)) where


import Data.Bits (Bits((.&.)))
import Data.Enum.Util (enumerate)
import Data.Set (Set)

import qualified Data.Set as Set

import Language.Java.Classfile.Extract (Extract)
import Language.Java.Classfile.Extractable (Extractable (extract))
import Data.Kind (Type)

-- | Using the 'FlagMask' instance of the type parameter, this will extract all the flags whose mask produced a non-zero value using '.&.'

newtype Flags a = Flags (Set a)
  deriving (Show)

instance (Extractable (FlagType a), Bounded a, Enum a, Ord a, FlagMask a, Bits (FlagType a), Num (FlagType a)) => Extractable (Flags a) where
  extract :: Extract (Flags a)
  extract = do
    flags <- extract @(FlagType a)
    let allFlags = enumerate @a
    let isContained flag = (maskOf flag .&. flags) /= 0

    pure . Flags . Set.fromList $ filter isContained allFlags

-- | Specify the masks and Parse type (size) a flag enum uses.

class FlagMask a where
  type FlagType a :: Type
  maskOf :: a -> FlagType a

