-- | Parse the Magic bytes at the beginning of every class-file.

{-# LANGUAGE InstanceSigs #-}
{-# LANGUAGE TypeFamilies #-}
{-# LANGUAGE DataKinds #-}
{-# LANGUAGE StandaloneKindSignatures #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE GADTs #-}
module Language.Java.Classfile.Magic (Magic(..)) where
import Data.Word (Word32)
import Language.Java.Classfile.Extractable (Extractable, extract, expectConstant)
import Language.Java.Classfile.Extract (Extract)
import Data.Hex (Hex (Hex))
import Language.Java.Classfile.Stage (Stage(Parse, Resolve))
import Data.Kind (Type)

-- | 'Stage'-indexed Magic type. The 'Resolve' stage is only a unit type.

type Magic :: Stage -> Type
data Magic stage where
  Magic :: Hex Word32 -> Magic Parse
  Cafebabe :: Magic Resolve

instance Extractable (Magic Parse) where
  extract :: Extract (Magic Parse)
  extract = Magic . Hex <$> expectConstant 0xCAFEBABE

-- >>> :set -XOverloadedLists
-- >>> import Language.Java.Classfile.Extract (runExtract)
-- >>> runExtract [0xCA, 0xFE, 0xBA, 0xBE] (Hex <$> (extract :: Extract (Word32)))
-- Right ("",Hex 0xcafebabe)

